Package: MsBackendMetaboLights
Title: Retrieve Mass Spectrometry Data from MetaboLights
Version: 1.7.4
Authors@R: 
    c(person(given = "Johannes", family = "Rainer",
             email = "Johannes.Rainer@eurac.edu",
             role = c("aut", "cre"),
             comment = c(ORCID = "0000-0002-6977-7147")),
      person(given = "Philippine", family = "Louail",
             email = "philippine.louail@eurac.edu",
             role = c("aut"),
             comment = c(ORCID = "0009-0007-5429-6846")),
      person(given = "Gabriele", family = "Tomè",
             email = "gabriele.tome@eurac.edu",
             role = c("aut"),
             comment = c(ORCID = "0000-0002-3976-6068",
             fnd = "MetaRbolomics4Galaxy project (CUP: D53C25001030003) co-funded by the Autonomous Province of Bolzano under the Joint Projects South Tyrol–Germany 2025 program."))
	     )
Description: MetaboLights is one of the main public repositories for storage
             of metabolomics experiments, which includes analysis results as
	     well as raw data. The MsBackendMetaboLights package provides
	     functionality to retrieve and represent mass spectrometry (MS)
	     data from MetaboLights. Data files are downloaded and cached
	     locally avoiding repetitive downloads. MS data from metabolomics
	     experiments can thus be directly and seamlessly integrated into
	     R-based analysis workflows with the Spectra and
	     MsBackendMetaboLights package.
Depends: R (>= 4.2.0), Spectra (>= 1.15.12)
Imports: curl, ProtGenerics, BiocFileCache, S4Vectors, methods,
        progress, utils, MsCoreUtils (>= 1.23.8)
Suggests: testthat, rmarkdown, mzR, knitr, BiocStyle
License: Artistic-2.0
Encoding: UTF-8
VignetteBuilder: knitr
BugReports: https://github.com/RforMassSpectrometry/MsBackendMetaboLights/issues
URL: https://github.com/RforMassSpectrometry/MsBackendMetaboLights
biocViews: Infrastructure, MassSpectrometry, Metabolomics, DataImport,
        Proteomics
Roxygen: list(markdown=TRUE)
RoxygenNote: 8.0.0
git_url: https://git.bioconductor.org/packages/MsBackendMetaboLights
git_branch: devel
git_last_commit: c998a0d
git_last_commit_date: 2026-07-14
Repository: Bioconductor 3.24
Date/Publication: 2026-07-22
NeedsCompilation: no
Packaged: 2026-07-22 23:05:13 UTC; biocbuild
Author: Johannes Rainer [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-6977-7147>),
  Philippine Louail [aut] (ORCID:
    <https://orcid.org/0009-0007-5429-6846>),
  Gabriele Tomè [aut] (ORCID: <https://orcid.org/0000-0002-3976-6068>,
    fnd: MetaRbolomics4Galaxy project (CUP: D53C25001030003) co-funded
    by the Autonomous Province of Bolzano under the Joint Projects
    South Tyrol–Germany 2025 program.)
Maintainer: Johannes Rainer <Johannes.Rainer@eurac.edu>

# Generated by roxygen2: do not edit by hand

export(MsBackendMetaboLights)
export(mtbls_assay_data)
export(mtbls_cached_data_files)
export(mtbls_delete_cache)
export(mtbls_ftp_path)
export(mtbls_list_files)
export(mtbls_metadata)
export(mtbls_sample_data)
export(mtbls_sync)
export(mtbls_sync_data_files)
exportClasses(MsBackendMetaboLights)
exportMethods(backendInitialize)
exportMethods(backendRequiredSpectraVariables)
importClassesFrom(Spectra,MsBackendDataFrame)
importClassesFrom(Spectra,MsBackendMzR)
importFrom(BiocFileCache,BiocFileCache)
importFrom(BiocFileCache,bfcinfo)
importFrom(BiocFileCache,bfcremove)
importFrom(BiocFileCache,bfcupdate)
importFrom(MsCoreUtils,retry)
importFrom(S4Vectors,DataFrame)
importFrom(Spectra,MsBackendMzR)
importFrom(curl,curl)
importFrom(curl,handle_setopt)
importFrom(curl,new_handle)
importFrom(methods,as)
importFrom(methods,callNextMethod)
importFrom(methods,new)
importFrom(methods,validObject)
importFrom(progress,progress_bar)
importFrom(utils,URLencode)
importFrom(utils,capture.output)
importFrom(utils,menu)
importFrom(utils,read.table)
importMethodsFrom(BiocFileCache,"bfcmeta<-")
importMethodsFrom(BiocFileCache,bfcmetalist)
importMethodsFrom(BiocFileCache,bfcquery)
importMethodsFrom(BiocFileCache,bfcrpath)
importMethodsFrom(ProtGenerics,backendInitialize)
importMethodsFrom(ProtGenerics,dataOrigin)
importMethodsFrom(Spectra,"[")
importMethodsFrom(Spectra,backendInitialize)
importMethodsFrom(Spectra,backendRequiredSpectraVariables)

# Retrieve Mass Spectrometry Data from MetaboLights

[![Project Status: Active – The project has reached a stable, usable state and is being actively developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![R-CMD-check-bioc](https://github.com/RforMassSpectrometry/MsBackendMetaboLights/workflows/R-CMD-check-bioc/badge.svg)](https://github.com/RforMassSpectrometry/MsBackendMetaboLights/actions?query=workflow%3AR-CMD-check-bioc)
[![codecov](https://codecov.io/gh/rformassspectrometry/MsBackendMetaboLights/graph/badge.svg?token=jpxt7OlA2k)](https://codecov.io/gh/rformassspectrometry/MsBackendMetaboLights)
[![:name status badge](https://rformassspectrometry.r-universe.dev/badges/:name)](https://rformassspectrometry.r-universe.dev/)
[![license](https://img.shields.io/badge/license-Artistic--2.0-brightgreen.svg)](https://opensource.org/licenses/Artistic-2.0)

This repository provides a *backend* for
[Spectra](https://github.com/RforMassSpectrometry/Spectra) objects that
represents and retrieves mass spectrometry (MS) data directly from metabolomics
experiments deposited at the public
[MetaboLights](https://www.ebi.ac.uk/metabolights/) repository. Mass
spectrometry data files of an experiment are downloaded and cached locally using
the [BiocFileCache](https://bioconductor.org/packages/BiocFileCache) package.


# Installation

The package can be installed with

```r
install.packages("BiocManager")
BiocManager::install("RforMassSpectrometry/MsBackendMetaboLights")
```


# Contributions

Contributions are highly welcome and should follow the [contribution
guidelines](https://rformassspectrometry.github.io/RforMassSpectrometry/articles/RforMassSpectrometry.html#contributions).
Also, please check the coding style guidelines in the [RforMassSpectrometry
vignette](https://rformassspectrometry.github.io/RforMassSpectrometry/articles/RforMassSpectrometry.html).

#' MetaboLights ID: MTBLS8735
#' ftp://ftp.ebi.ac.uk/pub/databases/metabolights/studies/public/MTBLS8735

library(testthat)
library(MsBackendMetaboLights)

test_check("MsBackendMetaboLights")

## Run tests with the unit test suite defined in the Spectra package to ensure
## compliance with the definitions of the MsBackend interface/class.
be <- backendInitialize(MsBackendMetaboLights(), mtblsId = "MTBLS39",
                        filePattern = "63A.cdf")

library(Spectra)
test_suite <- system.file("test_backends", "test_MsBackend",
                          package = "Spectra")
res <- test_dir(test_suite, stop_on_failure = TRUE)

## Cleanup
mtbls_delete_cache("MTBLS39")

################################################################################
## Utility functions for MetaboLights
##
################################################################################

#' @title Utility functions for the MetaboLights repository
#'
#' @name MetaboLights-utils
#'
#' @description
#'
#' [MetaboLights](https://www.ebi.ac.uk/metabolights/) is one of the main
#' public repositories for deposition of metabolomics experiments including
#' (raw) mass spectrometry (MS) and NMR data files and experimental/analysis
#' results. The experimental metadata and results are stored as plain text
#' files in ISA-tab format. Each MetaboLights experiment must provide a
#' file describing the samples analyzed and at least one *assay* file that
#' links between the experimental samples and the (raw and processed) data
#' files with quantification of metabolites/features in these samples.
#'
#' Each experiment in MetaboLights is identified with its unique identifier,
#' starting with *MTBLS* followed by a number. The data (metadata files and
#' MS/NMR data files) of an experiment are available through the repository's
#' ftp server.
#'
#' The functions listed here allow to query and retrieve information of a
#' data set/experiment from MetaboLights.
#'
#' - `mtbls_ftp_path()`: returns the FTP path for a provided MetaboLights ID.
#'   With `mustWork = TRUE` (the default) the function throws an error if
#'   the path is not accessible (either because the data set does not exist or
#'   no internet connection is available). The function returns a
#'   `character(1)` with the FTP path to the data set folder.
#'
#' - `mtbls_list_files()`: returns the available files (and directories) for the
#'   specified MetaboLights data set (i.e., the FTP directory content of the
#'   data set). The function returns a `character` vector with the relative
#'   file names to the absolute FTP path (`mtbls_ftp_path()`) of the data set.
#'   Parameter `pattern` allows to filter the file names and define which
#'   file names should be returned.
#'
#' - `mtbls_assay_data()`: retrieves one of the *assay* files for a
#'   MetaboLights data set (parameter `mtblsId`) returning its content as a
#'   `data.frame`. Parameter `assayName` allows to specify which assay file to
#'   load (if multiple are available).
#'
#' - `mtbls_sample_data()`: gets the *sample* file for a MetaboLights data set
#'   (parameter `mtblsId`) and returns its content as a `data.frame`.
#'
#' - `mtbls_metadata()`: gets one *assay* file for the specified MetaboLights
#'   data set (parameter `mtblsId`) and merges it with the respective *sample*
#'   information returning the content as a `data.frame`. Optional parameters
#'   `keepOntology`, `keepProtocol` and `simplify` allow to restrict the
#'   returned content to fewer columns.
#'
#' - `mtbls_cached_data_files()`: lists locally cached data files from
#'   MetaboLights. Since this function evaluates only local content it does not
#'   require an internet connection. With the default parameters all available
#'   data files are listed. The parameters can be used to restrict the lookup.
#'
#' - `mtbls_sync_data_files()`: synchronize data files of a specifies
#'   MetaboLights data set eventually downloading and locally caching them.
#'   Parameter `fileName` allows to specify names of selected data files to
#'   sync.
#'
#' - `mtbls_delete_cache()`: removes all local content for the MetaboLights
#'   data set with ID `mtblsId`. This will delete eventually present
#'   locally cached data files for the specified data set. This does not
#'   change any other data eventually present in the local `BiocFileCache`.
#'
#' @details
#'
#' File names are returned in **alphabetic order**.
#'
#' @param x `character(1)` with the ID of the MetaboLights data set (usually
#'     starting with a *MTBLS* followed by a number).
#'
#' @param mtblsId `character(1)` with the ID of a **single** MetaboLights data
#'     set/experiment.
#'
#' @param assayName `character` with the file names of assay files of the data
#'     set. If not provided (`assayName = character()`, the default), MS data
#'     files of all data set's assays are loaded. Use
#'     `mtbls_list_files(<MetaboLights ID>, pattern = "^a_")` to list all
#'     available assay files of a data set `<MetaboLights ID>`.
#'
#' @param mustWork for `mtbls_ftp_path()`: `logical(1)` whether the validity of
#'     the path should be verified or not. By default (with `mustWork = TRUE`)
#'     the function throws an error if either the data set does not exist or
#'     if the folder can not be accessed (e.g. if no internet connection is
#'     available).
#'
#' @param pattern for `mtbls_list_files()`, `mtbls_sync_data_files()` and
#'     `mtbls_cached_data_files()`: `character(1)` defining a pattern
#'     to filter the file names, such as `pattern = "^a_"` to retrieve the
#'     file names of all assay files of the data set (i.e., files with a name
#'     starting with `"a_"`). This parameter is passed to the [grepl()]
#'     function.
#'
#' @param fileName for `mtbls_sync_data_files()` and
#'     `mtbls_cached_data_files()`: optional `character`
#'     defining the names of specific data files of a data set that should be
#'     downloaded and cached.
#'
#' @param keepOntology for `mtbls_metadata()`: `logical(1)` whether to keep
#'     columns related to ontology. Default is `TRUE`.
#'
#' @param keepProtocol for `mtbls_metadata()`: `logical(1)` whether to keep
#'     columns with information related to protocols. Default is `TRUE`.
#'
#' @param simplify for `mtbls_metadata()`: `logical(1)` whether to simplify
#'     the result removing columns with only missing data or duplicated content.
#'     Default is `FALSE`.
#'
#' @return
#'
#' - For `mtbls_ftp_path()`: `character(1)` with the ftp path to the specified
#'   data set on the MetaboLights ftp server.
#' - For `mtbls_list_files()`: `character` with the names of the files in the
#'   data set's base ftp directory.
#' - For `mtbls_sync_data_files()` and `mtbls_cached_data_files()`: a
#'   `data.frame` with the MetaboLights ID, the assay name(s) and remote and
#'   local file names of the synchronized data files.
#'
#' @author Johannes Rainer, Philippine Louail
#'
#' @examples
#'
#' ## Get the FTP path to the data set MTBLS2
#' mtbls_ftp_path("MTBLS2")
#'
#' ## Retrieve available files (and directories) for the data set MTBLS2
#' mtbls_list_files("MTBLS2")
#'
#' ## Retrieve the available assay files (file names starting with "a_").
#' afiles <- mtbls_list_files("MTBLS2", pattern = "^a_")
#' afiles
#'
#' ## Read the content of one file. Connections to the MetaboLights ftp server
#' ## are limited and might fail, thus we use the `retry()` function to
#' ## retry on failure for 5 times (waiting `i * sleep_mult` seconds in between)
#' a <- MsCoreUtils::retry(
#'     read.table(paste0(mtbls_ftp_path("MTBLS2"), afiles[1L]),
#'     header = TRUE, sep = "\t", check.names = FALSE),
#'     ntimes = 5, sleep_mult = 4)
#' head(a)
#'
#' ## Get the assay information for one MTBLS data set
#' mtbls_assay_data("MTBLS2")
#'
#' ## Get the sample information for one data set
#' mtbls_sample_data("MTBLS2")
#'
#' ## List all available files
#' mtbls_cached_data_files()
NULL

#' @rdname MetaboLights-utils
#'
#' @export
mtbls_ftp_path <- function(x = character(), mustWork = TRUE) {
    if (length(x) != 1L)
        stop("'x' has to be a single ID.")
    res <- paste0("ftp://ftp.ebi.ac.uk/pub/databases/metabolights/",
                  "studies/public/", x, "/")
    if (mustWork)
        mtbls_list_files(x)
    res
}

#' @importFrom curl new_handle handle_setopt curl
#'
#' @importFrom MsCoreUtils retry
#'
#' @rdname MetaboLights-utils
#'
#' @export
mtbls_list_files <- function(x = character(), pattern = NULL) {
    cu <- new_handle()
    handle_setopt(cu, ftp_use_epsv = TRUE, dirlistonly = TRUE)
    tryCatch({
        con <- retry(
            curl(url = mtbls_ftp_path(x, mustWork = FALSE), "r", handle = cu),
            sleep_mult = .sleep_mult(), verbose = TRUE,
            retry_on = .RETRY_PATTERN)
    }, error = function(e) {
        stop("Failed to connect to MetaboLights. No internet connection? ",
             "Does the data set \"", x, "\" exist?\n - ", e$message,
             call. = FALSE)
    })
    fls <- retry(readLines(con), sleep_mult = .sleep_mult(),
                 retry_on = .RETRY_PATTERN, verbose = TRUE)
    close(con)
    if (length(pattern))
        fls[grepl(pattern, fls)]
    else fls
}

#' retrieves and reads the/all assay data file(s) for a given MetaboLights
#' data set and returns it/them as a `list` of `data.frame`s. The file
#' names of the respective assay data file(s) are reported as `names()` of
#' the returned `list`.
#'
#' @param x `character(1)` with the MetaboLights ID of the data set.
#'
#' @importFrom utils read.table
#'
#' @noRd
.mtbls_assay_list <- function(x = character()) {
    fpath <- mtbls_ftp_path(x, mustWork = FALSE)
    a_fls <- mtbls_list_files(x, pattern = "^a_")
    res <- lapply(a_fls, function(z) {
        retry(read.table(paste0(fpath, z),
                         sep = "\t", header = TRUE,
                         check.names = FALSE),
              sleep_mult = .sleep_mult(),
              retry_on = .RETRY_PATTERN, verbose = TRUE)
    })
    names(res) <- a_fls
    res
}

#' Extract the MS data files from MTBLS assay tables'
#' *Derived Spectral Data File* column(s). The function checks all present
#' columns in the provided `data.frame` and returns the content of the first of
#' these columns with files matching the provided `pattern`.
#'
#' @param x `data.frame` representing the content of an *assay* ISA file from
#'     MetaboLights.
#'
#' @param pattern `character(1)` with supported data types.
#'
#' @return `character` with the file names matching the provided pattern.
#'
#' @noRd
.mtbls_data_file_from_assay <-
    function(x, pattern = "mzML$|CDF$|cdf$|mzXML$",
             colname = "Derived Spectral Data File") {
        cls <- which(colnames(x) == colname)
        res <- character()
        for (i in cls) {
            keep <- grepl(pattern, x[[i]])
            if (any(keep)) {
                res <- x[[i]][keep]
                break
            }
        }
        res
    }

################################################################################
##
## File caching utils
##
################################################################################

#' @rdname MetaboLights-utils
#'
#' @export
mtbls_sync_data_files <- function(mtblsId = character(),
                                  assayName = character(),
                                  pattern = "mzML$|CDF$|cdf$|mzXML$",
                                  fileName = character()) {
    if (!length(mtblsId))
        stop("No MetaboLights data set ID provided with parameter 'mtblsId'")
    if (length(mtblsId) > 1)
        stop("Only a single MetaboLights ID can be submitted at a time")
    .mtbls_data_files(mtblsId, assayName, pattern, fileName)
}

#' @rdname MetaboLights-utils
#'
#' @export
mtbls_cached_data_files <- function(mtblsId = character(),
                                    assayName = character(),
                                    pattern = "*", fileName = character()) {
    res <- .mtbls_data_files_offline(mtblsId = mtblsId, assayName = assayName,
                                     pattern = pattern)
    if (length(fileName)) {
        keep <- basename(res$derived_spectral_data_file) %in% fileName
        if (!any(keep))
            stop("None of the 'fileName' found in data set \"", mtblsId, "\"")
        res <- res[keep, ]
    }
    res
}

.mtbls_fix_file_names <- function(mtblsId = character(),
                                  fileName = character()) {
    paste0(mtblsId, "_", fileName)
}

#' Get information on data files for a given MTBLS ID/assay eventually
#' downloading and caching them. This function needs an active internet
#' connection as it queries the MTBLS ftp server for available data files
#' that are then cached. The function returns the **local** file names
#' **from the cache**.
#'
#' The function:
#' - retrieves all "Derived Data Files" for all assays (or for specified assays)
#'   for one MetaboLights ID.
#' - uses BiocFileCache to cache these files, i.e. downloading them if they
#'   are not yet cached.
#' - returns a `data.frame` with all information.
#'
#' This `data.frame` has one row per data file with columns:
#' - `"rid"`: the BiocFileCache ID of each file.
#' - `"mtbls_id"`: the MTBLS ID
#' - `"mtbls_assay_name"`: the name of the assay file for each data file
#' - `"mtbls_assay_id"`: the ID of the assay file for each data file
#' - `"derived_spectral_data_file"`: the name of the data file in the assay
#'   file/table
#' - `"rpath"`: the name of the cached data file (full local path)
#'
#' @note
#'
#' Download from MsBackendMetaboLights is tried 3 times with an increasing time
#' delay between tries that can be configured using the
#' `"metabolights.sleep_mult"` option.
#'
#' @importFrom BiocFileCache BiocFileCache
#'
#' @importFrom progress progress_bar
#'
#' @importMethodsFrom BiocFileCache bfcrpath bfcmeta<-
#'
#' @importFrom utils capture.output URLencode
#'
#' @noRd
.mtbls_data_files <- function(mtblsId = character(), assayName = character(),
                              pattern = "mzML$|CDF$|mzXML$",
                              fileName = character()) {
    assays <- .mtbls_assay_list(mtblsId)
    assays <- assays[sort(names(assays))]
    anames <- names(assays)
    if (length(assayName)) {
        if (!all(assayName %in% anames))
            stop("Not all assay names defined with 'assayName' are available ",
                 "for ", mtblsId, ". Available assay names are: \n",
                 paste0(" - \"", anames, "\"", collapse = "\n"), call. = FALSE)
        assays <- assays[anames %in% assayName]
    }
    fpath <- mtbls_ftp_path(mtblsId, mustWork = FALSE)
    dfiles <- lapply(assays, .mtbls_data_file_from_assay, pattern = pattern)
    ffiles <- unlist(dfiles, use.names = FALSE)
    if (!length(ffiles)) {
        ## Failsafe; use evaluate also raw data file
        dfiles <- lapply(assays, .mtbls_data_file_from_assay, pattern = pattern,
                         colname = "Raw Spectral Data File")
        ffiles <- unlist(dfiles, use.names = FALSE)
        if (!length(ffiles))
            stop("No files matching the provided file pattern found for ",
                 "MetaboLights data set ", mtblsId, ".", call. = FALSE)
        else
            message("Used data files from the assay's column \"Raw Spectral ",
                    "Data File\" since none were available in column ",
                    "\"Derived Spectral Data File\".")
    }
    if (length(fileName)) {
        keep <- basename(ffiles) %in% fileName
        if (!any(keep))
            stop("None of the 'fileName' found in data set \"", mtblsId, "\"")
        ffiles <- ffiles[keep]
        dfiles <- lapply(dfiles, function(z) z[basename(z) %in% fileName])
    }
    ## Cache files
    bfc <- BiocFileCache()
    assay_idx <- as.integer(as.factor(rep(names(dfiles), lengths(dfiles))))
    lfiles <- .bfc_cache_files(URLencode(paste0(fpath, ffiles)), bfc,
                                paste0(mtblsId, "_", assay_idx))
    ## Remove 0 size files and re-run
    fsize <- file.size(lfiles)
    fsize[is.na(fsize)] <- 0
    unlink(lfiles[fsize == 0])
    if (any(fsize == 0)) {
        message("Resuming failed downloads")
        ## Remove from file cache
        b <- as.data.frame(bfcinfo(bfc))
        rid <- b$rid[b$rpath %in% lfiles[fsize == 0]]
        if (length(rid)) bfcremove(bfc, rids = rid)
        mis <- .bfc_cache_files(URLencode(paste0(fpath, ffiles[fsize == 0])),
                                bfc, paste0(mtblsId, "_", assay_idx))
        names(lfiles)[match(mis, lfiles)] <- names(mis)
    }
    ## Add and store metadata to the cached files
    mdata <- data.frame(
        rid = names(lfiles),
        mtbls_id = mtblsId,
        mtbls_assay_name = rep(names(dfiles), lengths(dfiles)),
        mtbls_assay_id = assay_idx,
        derived_spectral_data_file = unlist(dfiles, use.names = FALSE))
    bfcmeta(bfc, name = "MTBLS", overwrite = TRUE) <- mdata
    mdata$rpath <- lfiles
    mdata[order(mdata$rpath), , drop = FALSE]
}

#' Helper's helper to just run the caching operation on a provided set of
#' files.
#'
#' @importFrom BiocFileCache bfcupdate
#'
#' @noRd
.bfc_cache_files <- function(x, bfc, mtblsAssayId = "MTBLS_1") {
    if (length(mtblsAssayId))
        mtblsAssayId <- rep(mtblsAssayId, length(x))
    pb <- progress_bar$new(format = paste0("[:bar] :current/:",
                                           "total (:percent) in ",
                                           ":elapsed"),
                           total = length(x), clear = FALSE)
    unlist(lapply(1:length(x), function(i) {
        pb$tick()
        invisible(capture.output(suppressMessages(
            f <- retry(bfcrpath(bfc, x[[i]], fname = "exact"),
                       sleep_mult = .sleep_mult(), verbose = TRUE,
                       retry_on = .RETRY_PATTERN, warningsAsErrors = TRUE))))
        if (length(f) && !grepl(paste0("^", mtblsAssayId[[i]]), basename(f))) {
            ## Rename file name to <MTBLS>_<ASSAY_ID>_<FILENAME> to avoid
            ## overwriting files from other data sets with same file name
            fnew <- file.path(dirname(f),
                              paste0(mtblsAssayId[[i]], "_", basename(f)))
            names(fnew) <- names(f)
            file.rename(f, fnew)
            suppressWarnings(
                bfcupdate(bfc, rids = names(f), rpath = fnew,
                          ask = FALSE, progress = FALSE))
            f <- fnew
        }
        f
    }))
}

#' Check for a given MTBLS ID and assay IDs/file names if we have cached data
#' files. This function is supposed to work also offline using only previously
#' cached content. In contrast to `.mtbls_data_files()`, this function just
#' queries the BiocFileCache for content and returns a `data.frame` with
#' all cached data files for a given MTBLS ID, assay name and pattern. The
#' returned `data.frame` has the same format as the one returned by
#' `.mtbls_data_files()`.
#'
#' @importMethodsFrom BiocFileCache bfcquery
#'
#' @noRd
.mtbls_data_files_offline <- function(mtblsId = character(),
                                      assayName = character(),
                                      pattern = "mzML$|CDF$|mzXML$") {
    bfc <- BiocFileCache()
    if (!.mtbls_has_mtbls_table())
        stop("No local MetaboLights cache available. Please re-run with ",
             "'offline = FALSE' first.", call. = FALSE)
    res <- as.data.frame(bfcquery(bfc, mtblsId, field = "mtbls_id"))
    if (length(assayName)) {
        res <- res[res$mtbls_assay_name %in% assayName, ]
    }
    res <- res[grepl(pattern, res$derived_spectral_data_file), ]
    if (!nrow(res))
        stop("No locally cached data files found for the specified ",
             "parameters.", call. = FALSE)
    res <- res[, c("rid", "mtbls_id", "mtbls_assay_name", "mtbls_assay_id",
                   "derived_spectral_data_file", "rpath")]
    res[order(res$rpath), , drop = FALSE]
}

#' @importMethodsFrom BiocFileCache bfcmetalist
#'
#' @noRd
.mtbls_has_mtbls_table <- function() {
    bfc <- BiocFileCache()
    any(bfcmetalist(bfc) == "MTBLS")
}

#' @rdname MetaboLights-utils
#'
#' @importFrom BiocFileCache bfcremove bfcinfo
#'
#' @export
mtbls_delete_cache <- function(mtblsId = character()) {
    bfc <- BiocFileCache()
    b <- as.data.frame(bfcinfo(bfc))
    if (nrow(b) && any(colnames(b) == "mtbls_id")) {
        if (length(mtblsId)) {
            rem <- b[b$mtbls_id %in% mtblsId, ]
            bfcremove(bfc, rids = rem$rid)
        }
    }
}

#' @rdname MetaboLights-utils
#'
#' @importFrom utils menu
#'
#' @export
mtbls_assay_data <- function(mtblsId = character(), assayName = character()) {
    if (!length(mtblsId))
        return(data.frame())
    all_fls <- mtbls_list_files(mtblsId)
    assays <- grep("^a", all_fls, value = TRUE)
    if (length(assayName)) {
        if (any(assayName %in% assays))
            assays <- assayName
        else stop("Specified assay \"", assayName, "\" does not exist.",
                  call. = FALSE)
    }
    if (length(assays) > 1) {
        message("Multiple assay files found:\n")
        sel <- menu(assays, title = "Please choose assay file:")
        assays <- assays[sel]
    }
    fpath <- mtbls_ftp_path(mtblsId, mustWork = FALSE)
    retry(read.table(paste0(fpath, assays), header = TRUE, sep = "\t",
                     check.names = FALSE, comment.char = "", quote = ""),
          sleep_mult = .sleep_mult(), retry_on = .RETRY_PATTERN,
          verbose = TRUE)
}

#' @rdname MetaboLights-utils
#'
#' @export
mtbls_sample_data <- function(mtblsId = character()) {
    if (!length(mtblsId))
        return(data.frame())
    all_fls <- mtbls_list_files(mtblsId)
    x <- grep("^s", all_fls, value = TRUE)
    fpath <- mtbls_ftp_path(mtblsId, mustWork = FALSE)
    retry(read.table(paste0(fpath, x), header = TRUE, sep = "\t",
                     check.names = FALSE, comment.char = "", quote = ""),
          sleep_mult = .sleep_mult(), retry_on = .RETRY_PATTERN,
          verbose = TRUE)
}

#' @rdname MetaboLights-utils
#'
#' @export
mtbls_metadata <- function(mtblsId = character(), assayName = character(),
                           keepOntology = TRUE, keepProtocol = TRUE,
                           simplify = FALSE) {
    if (!length(mtblsId))
        return(data.frame())
    a <- mtbls_assay_data(mtblsId, assayName)
    s <- mtbls_sample_data(mtblsId)
    o <- match(a$`Sample Name`, s$`Sample Name`)
    m <- cbind(a, s[o, , drop = FALSE])
    if (!keepOntology || !keepProtocol || simplify)
        m <- .clean_merged(m, keepOntology = keepOntology,
                           keepProtocol = keepProtocol, simplify = simplify)
    m
}

#' Function that takes the extra parameters and clean the metadata if asked by
#' the user.
#'
#' @noRd
.clean_merged <- function(x, keepProtocol, keepOntology, simplify) {
    # remove ontology
    if (!keepOntology)
        x <- x[, -which(grepl("Term", names(x))), drop = FALSE]
    # remove protocol
    if (!keepProtocol)
        x <- x[, -which(grepl("Protocol|Parameter", names(x))),  drop = FALSE]
    # remove duplicated columns contents and NAs
    if (simplify) {
        x <- x[, !duplicated(as.list(x)), drop = FALSE]
        x <- x[, colSums(is.na(x)) != nrow(x), drop = FALSE]
    }
    x
}

.sleep_mult <- function() {
    as.integer(getOption("metabolights.sleep_mult", default = 7L))
}

## "download" download failed
## "imeout"   Timeout was reached
## "request"  Failed to performm HTTP request.
.RETRY_PATTERN <- "download|imeout|request"

alist_ms <- .mtbls_assay_list("MTBLS2")
Sys.sleep(4)
alist_nmr <- .mtbls_assay_list("MTBLS123")

test_that("mtbls_ftp_path works", {
    res <- mtbls_ftp_path("A", mustWork = FALSE)
    expect_true(grepl("^ftp://", res))
    expect_true(grepl("A/$", res))

    expect_error(mtbls_ftp_path("A", mustWork = TRUE), "Failed to connect")

    res <- mtbls_ftp_path("MTBLS1")
    expect_true(grepl("MTBLS1/$", res))

    expect_error(mtbls_ftp_path(c("A", "B")), "single ID")
})

test_that("mtbls_list_files works", {
    Sys.sleep(4)
    res <- mtbls_list_files("MTBLS8735", pattern = "^a_")
    expect_true(length(res) == 2)
    expect_error(mtbls_list_files("AAA"), "Failed to connect")
})

test_that(".mtbls_assay_list works", {
    Sys.sleep(4)
    res <- .mtbls_assay_list("MTBLS8735")
    expect_true(is.list(res))
    expect_true(length(res) == 2L)
    expect_true(is.data.frame(res[[1L]]))
})

test_that(".mtbls_data_file_from_assay works", {
    res <- .mtbls_data_file_from_assay(alist_ms[[1L]])
    expect_true(is.character(res))
    expect_true(length(res) == 16)

    ## check the second column containing mzData files
    res <- .mtbls_data_file_from_assay(alist_ms[[1L]], pattern = "mzData")
    expect_true(is.character(res))
    expect_true(length(res) == 16)

    res <- .mtbls_data_file_from_assay(alist_nmr[[1L]])
    expect_true(is.character(res))
    expect_true(length(res) == 0)
})

test_that(".mtbls_data_files and .mtbls_data_files_offline works", {
    ## error
    expect_error(.mtbls_data_files(mtblsId = "MTBLS2",
                                   assayName = "does not exist"),
                 "Not all assay names")
    expect_error(.mtbls_data_files(mtblsId = "MTBLS100"), "No files matching")

    mtbls_delete_cache("MTBLS39")
    ## bfc <- BiocFileCache::BiocFileCache()
    ## BiocFileCache::cleanbfc(bfc, days = -10, ask = FALSE)
    ## BiocFileCache::bfcmetaremove(bfc, "MTBLS")

    ## Error if no cache available
    with_mocked_bindings(
        ".mtbls_has_mtbls_table" = function() FALSE,
        code = expect_error(.mtbls_data_files_offline("MTBLS39"),
                            "No local MetaboLights cache")
    )

    ## Cache the data: MTBLS39 contains small cdf files, but they are listed
    ## in the Raw Spectral Data File column. Will use a specfic pattern to
    ## just load 3 files.
    a <- .mtbls_data_files("MTBLS39", pattern = "63A.cdf")
    expect_true(is.data.frame(a))
    expect_true(nrow(a) == 3)
    expect_true(all(a$mtbls_id == "MTBLS39"))
    ## Re-call function the data.
    Sys.sleep(4)
    b <- .mtbls_data_files("MTBLS39", pattern = "63A.cdf")
    expect_true(is.data.frame(b))
    expect_true(nrow(b) == 3)
    expect_true(all(b$mtbls_id == "MTBLS39"))
    expect_equal(a$rpath, b$rpath)

    ## with fileNames
    expect_error(.mtbls_data_files("MTBLS39", pattern = "63A.cdf",
                                   fileName = c("a", "b")), "None of the ")

    ## with assayName
    Sys.sleep(4)
    b <- .mtbls_data_files(
        "MTBLS39", pattern = "63A.cdf",
        assayName = paste0("a_MTBLS39_the_plasticity_of_the_grapevine_berry",
                           "_transcriptome_metabolite_profiling_mass",
                           "_spectrometry.txt"))
    expect_true(is.data.frame(b))
    expect_true(nrow(b) == 3)
    expect_true(all(b$mtbls_id == "MTBLS39"))
    expect_equal(a$rpath, b$rpath)

    expect_true(.mtbls_has_mtbls_table())

    ## Use offline
    expect_error(.mtbls_data_files_offline("MTBLS39", assayName = "something"),
                 "No locally cached data files")

    d <- .mtbls_data_files_offline("MTBLS39", pattern = "63A.cdf")
    expect_true(is.data.frame(a))
    expect_true(nrow(a) == 3)
    expect_true(all(a$mtbls_id == "MTBLS39"))
    expect_equal(a$rpath, d$rpath)
})

test_that("mtbls_sync_data_files works", {
    expect_error(mtbls_sync_data_files(), "No MetaboLights data")
    expect_error(mtbls_sync_data_files(c("a", "b")), "single")
    res <- mtbls_sync_data_files("MTBLS39", pattern = "*",
                                 fileName = c("AM063A.cdf"))
    expect_true(is.data.frame(res))
    expect_equal(nrow(res), 1L)
    expect_equal(res$mtbls_id, "MTBLS39")
})

test_that("mtbls_cached_data_files works", {
    res <- mtbls_cached_data_files()
    expect_true(is.data.frame(res))
    expect_true(nrow(res) > 0)

    expect_error(mtbls_cached_data_files(fileName = "otehr"), "None of the ")

    res <- mtbls_cached_data_files(fileName = "AM063A.cdf")
    expect_true(is.data.frame(res))
    expect_equal(basename(res$derived_spectral_data_file), "AM063A.cdf")
})


test_that("mtbls_delete_cache works", {
    bfc <- BiocFileCache()
    l <- length(bfc)
    mtbls_delete_cache()
    expect_equal(length(bfc), l)

    mtbls_delete_cache("MTBLS39")
    i <- bfcinfo(bfc)
    expect_true(!any(i$mtbls_id %in% "MTBLS39"))
})

test_that("mtbls_assay_data works", {
    id <- "MTBLS2"
    res <- mtbls_assay_data()
    expect_true(is.data.frame(res))
    expect_true(nrow(res) == 0L)

    expect_error(mtbls_assay_data(id, "aaaa"), "does not exist")
    res <- mtbls_assay_data(id)
    expect_true(is.data.frame(res))
    expect_true(nrow(res) > 0)
})

test_that("mtbls_sample_data works", {
    id <- "MTBLS2"
    res <- mtbls_sample_data()
    expect_true(is.data.frame(res))
    expect_true(nrow(res) == 0L)

    res <- mtbls_sample_data(id)
    expect_true(is.data.frame(res))
    expect_true(nrow(res) > 0)
})

test_that("mtbls_metadata works", {
    res <- mtbls_metadata()
    expect_true(is.data.frame(res))
    expect_true(nrow(res) == 0L)

    res <- mtbls_metadata("MTBLS2")
    expect_true(is.data.frame(res))
    expect_true(nrow(res) > 0)
    res_2 <- mtbls_metadata("MTBLS2", keepProtocol = FALSE)
    expect_true(ncol(res) > ncol(res_2))
})

test_that(".clean_merged function works correctly", {
    tbc <- data.frame(
        Protocol_A = c(1, 2, 3),
        Term_B = c("ontology1", "ontology2", "ontology3"),
        Parameter_C = c(10, 20, 30),
        Term_D = c("ontology1", "ontology2", "ontology3"),
        Data_E = c(NA, NA, NA),
        Duplicate_F = c(1, 2, 3),
        stringsAsFactors = FALSE
    )
    result <- .clean_merged(tbc, keepProtocol = TRUE,
                            keepOntology = TRUE,
                            simplify = FALSE)
    expect_equal(names(result), names(tbc))

    result <- .clean_merged(tbc, keepProtocol = TRUE,
                            keepOntology = FALSE, simplify = FALSE)
    expect_equal(names(result), c("Protocol_A", "Parameter_C", "Data_E",
                                  "Duplicate_F"))

    result <- .clean_merged(tbc, keepProtocol = FALSE,
                            keepOntology = TRUE, simplify = FALSE)
    expect_equal(names(result), c("Term_B", "Term_D", "Data_E", "Duplicate_F"))

    result <- .clean_merged(tbc, keepProtocol = FALSE, keepOntology = FALSE,
                            simplify = FALSE)
    expect_equal(names(result), c("Data_E", "Duplicate_F"))

    result <- .clean_merged(tbc, keepProtocol = TRUE, keepOntology = TRUE,
                            simplify = TRUE)
    expect_equal(names(result), c("Protocol_A", "Term_B", "Parameter_C"))


    result <- .clean_merged(tbc, keepProtocol = FALSE, keepOntology = FALSE,
                            simplify = TRUE)
    expect_equal(names(result), "Duplicate_F")
})

test_that(".sleep_mult works", {
    expect_equal(.sleep_mult(), 7L)
})

test_that(".bfc_cache_files works", {
    bfc <- BiocFileCache()
    p <- "ftp://ftp.ebi.ac.uk/pub/databases/metabolights/studies/public/MTBLS39/"
    fls <- c("FILES/CS073B.cdf", "FILES/MN063A.cdf")
    res <- .bfc_cache_files(paste0(p, fls), bfc)
    expect_true(all(file.exists(res)))
    ## Clean up
    tmp <- bfcinfo(bfc)
    ids <- tmp$rid[tmp$rpath %in% res]
    bfcremove(bfc, rids = ids)
})

#' @title MsBackend representing MS data from MetaboLights
#'
#' @name MsBackendMetaboLights
#'
#' @aliases MsBackendMetaboLights-class
#'
#' @description
#'
#' `MsBackendMetaboLights` retrieves and represents mass spectrometry (MS)
#' data from metabolomics experiments stored in the
#' [MetaboLights](https://www.ebi.ac.uk/metabolights/) repository. The backend
#' directly extends the [Spectra::MsBackendMzR] backend from the *Spectra*
#' package and hence supports MS data in mzML, netCDF and mzXML format. Data
#' in other formats can not be loaded with `MsBackendMetaboLights`.
#' Upon initialization with the `backendInitialize()` method, the
#' `MsBackendMetaboLights` backend downloads and caches the MS data files of
#' an experiment locally avoiding hence repeated download of the data. Data
#' files are listed and returned in **alphabetic order**.
#' The local data cache is managed by Bioconductor's *BiocFileCache* package.
#' See the help and vignettes from that package for details on cached data
#' resources. Additional utility function for management of cached files are
#' also provided by *MsBackendMetaboLights*. See help for
#' [mtbls_cached_data_files()] for more information.
#'
#' @section Initialization and loading of data:
#'
#' New instances of the class can be created with the `MsBackendMetaboLights()`
#' function. Data is loaded and initialized using the `backendInitialize()`
#' function which can be configured with parameters `mtblsId`, `assayName` and
#' `filePattern`. `mtblsId` must be the ID of a **single** (existing)
#' MetaboLights data set. Parameter `assayName` allows to define specific
#' *assays* of the MetaboLights data set from which the data files should be
#' loaded. If provided, it should be the file name(s) of the respective
#' assay(s) in MetaboLights (use e.g.
#' `mtbls_list_files(<MetaboLights ID>, pattern = "^a_")` to list all available
#' assay files for a given MetaboLights ID `<MetaboLights ID>`). By default,
#' with `assayName = character()` MS data files from **all** assays of a data
#' set are loaded. Optional parameter `filePattern` defines the pattern that
#' should be used to filter the file names of the MS data files. It defaults
#' to data files with file endings of supported MS data files. The optional
#' `fileName` parameter allows to further restrict the loaded data to specific
#' data file(s), defined by their base file name(s).
#' `backendInitialize()` requires an active internet connection as the
#' function first compares the remote file content to the locally cached files
#' and eventually synchronizes changes/updates. This can be skipped with
#' `offline = TRUE` in which case only locally cached content is queried.
#'
#' The `backendRequiredSpectraVariables()` function returns the names of the
#' spectra variables required for the backend to provide the MS data.
#'
#' The `mtbls_sync()` function can be used to *synchronize* the local data
#' cache and ensure that all data files are locally available. The function
#' will check the local cache and eventually download missing data files from
#' the MetaboLights repository.
#'
#' @note
#'
#' To account for high server load and eventually failing or rejected
#' downloads from the MetaboLights ftp server, the download functions
#' repeatedly retry to download a file. An error is thrown if download fails
#' for 3 consecutive attempts. Between each attemp, the function waits
#' for an increasing time period (5 seconds between the first and second
#' and 10 seconds between the 2nd and 3rd attempt). This time period can
#' also be configured with the `"metabolights.sleep_mult"` option, which
#' defines the *sleep time multiplicator* (defaults to 5).
#'
#' @param object an instance of `MsBackendMetaboLights`.
#'
#' @param mtblsId `character(1)` with the ID of a single MetaboLights data
#'     set/experiment.
#'
#' @param assayName `character` with the file names of assay files of the data
#'     set. If not provided (`assayName = character()`, the default), MS data
#'     files of all data set's assays are loaded. Use
#'     `mtbls_list_files(<MetaboLights ID>, pattern = "^a_")` to list all
#'     available assay files of a data set `<MetaboLights ID>`.
#'
#' @param filePattern `character` with the pattern defining the supported (or
#'     requested) file types. Defaults to
#'     `filePattern = "mzML$|CDF$|cdf$|mzXML$"` hence restricting to mzML,
#'     CDF and mzXML files which are supported by *Spectra*'s
#'     `MsBackendMzR` backend.
#'
#' @param fileName `character` with optional base file name(s) of selected
#'     data files to load. If provided, only data files whose base file name
#'     matches one of `fileName` are loaded. Defaults to
#'     `fileName = character()` in which case all matching data files are
#'     loaded. An error is thrown if none of the `fileName` is found in the
#'     data set.
#'
#' @param offline `logical(1)` whether only locally cached content should be
#'     evaluated/loaded.
#'
#' @param x an instance of `MsBackendMetaboLights`.
#'
#' @param ... additional parameters; currently ignored.
#'
#' @return
#'
#' - For `MsBackendMetaboLights()`: an instance of `MsBackendMetaboLights`.
#' - For `backendInitialize()`: an instance of `MsBackendMetaboLights` with
#'   the MS data of the specified MetaboLights data set.
#' - For `backendRequiredSpectraVariables()`: `character` with spectra
#'   variables that are needed for the backend to provide the MS data.
#' - For `mtbls_sync()`: the input `MsBackendMetaboLights` with the paths to
#'   the locally cached data files being eventually updated.
#'
#' @details
#'
#' File names for data files are by default extracted from the column
#' `"Derived Spectral Data File"` of the MetaboLights data set's *assay*
#' table. If this column does not contain any supported file names, the
#' assay's column `"Raw Spectral Data File"` is evaluated instead.
#'
#' The backend uses the
#' [BiocFileCache](https://bioconductor.org/packages/BiocFileCache) package for
#' caching of the data files. These are stored in the default local
#' *BiocFileCache* cache along with additional metadata that includes the
#' MetaboLights ID and the assay file name with which the data file is
#' associated with. Note that at present only MS data files in *mzML*, *CDF*
#' and *mzXML* format are supported.
#'
#' The `MsBackendMetaboLights` backend defines and provides additional spectra
#' variables `"mtbls_id"`, `"mtbls_assay_name"` , `"mtbls_assay_id"` and
#' `"derived_spectral_data_file"` that list the MetaboLights ID, the name of
#' the assay file and the original data file name on the MetaboLights ftp
#' server for each individual spectrum. The `"derived_spectral_data_file"` can
#' be used for the mapping between the experiment's samples and the
#' individual data files, respective their spectra. This mapping is provided
#' in the MetaboLights assay file.
#'
#' The `MsBackendMetaboLights` backend is considered *read-only* and does
#' thus not support changing *m/z* and intensity values directly.
#'
#' @importClassesFrom Spectra MsBackendMzR
#'
#' @importClassesFrom Spectra MsBackendDataFrame
#'
#' @importFrom S4Vectors DataFrame
#'
#' @exportClass MsBackendMetaboLights
#'
#' @author Philippine Louail, Johannes Rainer
#'
#' @examples
#'
#' library(MsBackendMetaboLights)
#'
#' ## List files of a MetaboLights data set
#' mtbls_list_files("MTBLS39")
#'
#' ## Initialize a MsBackendMetaboLights representing all MS data files of
#' ## the data set with the ID "MTBLS39". This will download and cache all
#' ## files and subsequently load and represent them in R.
#'
#' be <- backendInitialize(MsBackendMetaboLights(), "MTBLS39")
#' be
#'
#' ## The `mtbls_sync()` function can be used to ensure that all data files are
#' ## available locally. This function will eventually download missing data
#' ## files or update their paths.
#' be <- mtbls_sync(be)
NULL

setClass("MsBackendMetaboLights",
         contains = "MsBackendMzR")

#' @rdname MsBackendMetaboLights
#'
#' @importFrom methods new
#'
#' @export
MsBackendMetaboLights <- function() {
    new("MsBackendMetaboLights")
}

#' @rdname MsBackendMetaboLights
#'
#' @importMethodsFrom ProtGenerics backendInitialize
#'
#' @importMethodsFrom Spectra backendInitialize
#'
#' @importMethodsFrom Spectra [
#'
#' @importMethodsFrom ProtGenerics dataOrigin
#'
#' @importFrom methods callNextMethod
#'
#' @importFrom methods as
#'
#' @importFrom Spectra MsBackendMzR
#'
#' @exportMethod backendInitialize
setMethod(
    "backendInitialize", "MsBackendMetaboLights",
    function(object, mtblsId = character(), assayName = character(),
             filePattern = "mzML$|CDF$|cdf$|mzXML$", fileName = character(),
             offline = FALSE, ...) {
        dots <- list(...)
        if (any(names(dots) == "data"))
            stop("Parameter 'data' is not supported for ",
                 "'MsBackendMetaboLights'. A 'MsBackendMetaboLights' object ",
                 "can only be instantiated using 'backendInitialize()'.")
        if (length(mtblsId) != 1)
            stop("Parameter 'mtblsId' is required and can only be a single ID ",
                 "of a MetaboLights data set.")
        if (offline)
            mdata <- mtbls_cached_data_files(mtblsId, assayName, filePattern,
                                             fileName)
        else mdata <- mtbls_sync_data_files(mtblsId, assayName, filePattern,
                                            fileName)
        object <- backendInitialize(MsBackendMzR(), files = mdata$rpath)
        idx <- match(dataOrigin(object),
                     normalizePath(mdata$rpath, mustWork = FALSE))
        object@spectraData$mtbls_id <- mdata$mtbls_id[idx]
        object@spectraData$mtbls_assay_name <- mdata$mtbls_assay_name[idx]
        object@spectraData$mtbls_assay_id <- mdata$mtbls_assay_id[idx]
        object@spectraData$derived_spectral_data_file <-
            mdata$derived_spectral_data_file[idx]
        object <- as(object, "MsBackendMetaboLights")
    })

#' @rdname MsBackendMetaboLights
#'
#' @importMethodsFrom Spectra backendRequiredSpectraVariables
#'
#' @exportMethod backendRequiredSpectraVariables
setMethod(
    "backendRequiredSpectraVariables", "MsBackendMetaboLights",
    function(object, ...) {
        c(callNextMethod(), "mtbls_id", "mtbls_assay_name", "mtbls_assay_id",
          "derived_spectral_data_file")
    })

.valid_mtbls_required_columns <- function(object) {
    if (nrow(object@spectraData)) {
        if (!all(c("mtbls_id", "mtbls_assay_name", "mtbls_assay_id",
                   "derived_spectral_data_file") %in%
                 colnames(object@spectraData)))
            return(paste0("One or more of required spectra variable(s) ",
                          "\"mtbls_id\", \"mtbls_assay_name\",
                          \"mtbls_assay_id\", \"derived_spectral_data_file\"",
                          "is (are) missing"))
    }
    character()
}

.valid_files_local <- function(object) {
    if (nrow(object@spectraData)) {
        if (!all(file.exists(object@spectraData$dataStorage)))
            return(paste0("One or more of the data files are not found in ",
                          "the local cache. Please run `mtbls_sync()` on ",
                          "the data object."))
    }
    character()
}

setValidity("MsBackendMetaboLights", function(object) {
    msg <- .valid_mtbls_required_columns(object)
    msg <- c(msg, .valid_files_local(object))
    if (length(msg)) return(msg)
    else TRUE
})

#' @importFrom methods validObject
#'
#' @rdname MsBackendMetaboLights
#'
#' @export
mtbls_sync <- function(x, offline = FALSE) {
    if (!inherits(x, "MsBackendMetaboLights"))
        stop("'x' is expected to be an instance of 'MsBackendMetaboLights'")
    sdata <- unique(
        as.data.frame(x@spectraData[, c("mtbls_id", "mtbls_assay_name",
                            "mtbls_assay_id", "derived_spectral_data_file")]))
    cn <- c("derived_spectral_data_file", "rpath")
    res <- lapply(split(sdata, sdata$mtbls_id), function(z, offline) {
        if (offline)
            mtbls_cached_data_files(
                sdata$mtbls_id[1L], pattern = "*",
                fileName = basename(sdata$derived_spectral_data_file))[, cn]
        else
            mtbls_sync_data_files(
                 sdata$mtbls_id[1L], pattern = "*",
                 fileName = basename(sdata$derived_spectral_data_file))[, cn]
    }, offline = offline)
    res <- do.call(rbind, res)
    if (!all(sdata$derived_spectral_data_file %in%
        res$derived_spectral_data_file))
        stop("Some of the data files are not available. Please run with ",
             "'offline = FALSE' to ensure data missing data files get ",
             "downloaded.")
    x@spectraData$dataStorage <- res[match(
        x@spectraData$derived_spectral_data_file,
        res$derived_spectral_data_file), "rpath"]
    validObject(x)
    x
}

## MTBLS10555: lists files in assay column that don't exist. Will result in an
##             error when we try to download the data. The data set contains
##             small data files. maybe use pattern sham-1-10.mzML
## MTBLS39: cdf files listed in Raw Spectral Data File column. Maybe use a
##          specific pattern to load/cache only some files.
## MTBLS243: mzML.gz files. Can eventually use an additional filePattern?

test_that("MsBackendMetaboLights works", {
    res <- MsBackendMetaboLights()
    expect_s4_class(res, "MsBackendMetaboLights")
    expect_true(inherits(res, "MsBackendMzR"))
})

test_that("backendInitialize,MsBackendMetaboLights works", {
    ## Test errors
    expect_error(backendInitialize(MsBackendMetaboLights(),
                                   data = data.frame(a = 3)),
                 "Parameter 'data' is not supported")
    expect_error(backendInitialize(MsBackendMetaboLights(),
                                   mtblsId = c("a", "b")),
                 "Parameter 'mtblsId' is required and can")
    expect_error(backendInitialize(MsBackendMetaboLights(), mtblsId = "a"),
                 "Failed to connect")

    ## Test NMR data set
    expect_error(backendInitialize(MsBackendMetaboLights(),
                                   mtblsId = "MTBLS100"), "No files matching")
    ## Test real data set.
    res <- backendInitialize(MsBackendMetaboLights(), mtblsId = "MTBLS39",
                             filePattern = "63A.cdf")
    expect_s4_class(res, "MsBackendMetaboLights")
    expect_true(all(c("mtbls_id", "mtbls_assay_name", "mtbls_assay_id",
                      "derived_spectral_data_file") %in%
                    Spectra::spectraVariables(res)))
    expect_true(all(res$mtbls_id == "MTBLS39"))

    ## Offline
    res_o <- backendInitialize(MsBackendMetaboLights(), mtblsId = "MTBLS39",
                               filePattern = "63A.cdf", offline = TRUE)
    expect_equal(Spectra::rtime(res), Spectra::rtime(res_o))

    Sys.sleep(4)
    res_f <- backendInitialize(MsBackendMetaboLights(), mtblsId = "MTBLS39",
                               filePattern = "63A.cdf",
                               fileName = "AM063A.cdf")
    expect_s4_class(res_f, "MsBackendMetaboLights")
    expect_equal(unique(basename(res_f$derived_spectral_data_file)),
                 "AM063A.cdf")

    res_fo <- backendInitialize(MsBackendMetaboLights(), mtblsId = "MTBLS39",
                                filePattern = "63A.cdf",
                                fileName = "AM063A.cdf", offline = TRUE)
    expect_equal(Spectra::rtime(res_f), Spectra::rtime(res_fo))

    expect_error(
        backendInitialize(MsBackendMetaboLights(), mtblsId = "MTBLS39",
                          filePattern = "63A.cdf", fileName = "does_not_exist"),
        "None of the ")
    expect_error(
        backendInitialize(MsBackendMetaboLights(), mtblsId = "MTBLS39",
                          filePattern = "63A.cdf", fileName = "does_not_exist",
                          offline = TRUE),
        "None of the ")
})

test_that("backendRequiredSpectraVariables,MsBackendMetaboLights works", {
    expect_equal(backendRequiredSpectraVariables(MsBackendMetaboLights()),
                 c("dataStorage", "scanIndex", "mtbls_id", "mtbls_assay_name",
                   "mtbls_assay_id", "derived_spectral_data_file"))
})

test_that("mtbls_sync works", {
    expect_error(mtbls_sync(3, offline = TRUE), "'x' is expected to be")

    x <- backendInitialize(MsBackendMetaboLights(), mtblsId = "MTBLS39",
                           filePattern = "63A.cdf", offline = TRUE)
    res <- mtbls_sync(x, offline = TRUE)
    expect_equal(rtime(x), rtime(res))
    expect_equal(mz(x[1:50]), mz(res[1:50]))

    ## Remove local content.
    mtbls_delete_cache("MTBLS39")
    expect_error(mtbls_sync(x, offline = TRUE), "No locally cached data files")

    Sys.sleep(4)

    ## Re-add content
    res <- mtbls_sync(x, offline = FALSE)
    expect_equal(rtime(x), rtime(res))
    expect_equal(mz(x[1:50]), mz(res[1:50]))

    ## Error.
    with_mocked_bindings(
        "mtbls_cached_data_files" = function(mtblsId, ...) {
            data.frame(rid = c("1", "2"),
                       derived_spectral_data_file = c("a", "b"),
                       rpath = "tmp")
        },
        code = expect_error(mtbls_sync(x, offline = TRUE), "not available")
    )
})

test_that(".valid_mtbls_required_columns works", {
    x <- MsBackendMetaboLights()
    expect_equal(.valid_mtbls_required_columns(x), character())
    x@spectraData <- DataFrame(a = 1:4, b = "c")
    expect_match(.valid_mtbls_required_columns(x), "One or more")
    x@spectraData$mtbls_id <- 3
    x@spectraData$mtbls_assay_name <- "a"
    x@spectraData$mtbls_assay_id <- 1
    x@spectraData$derived_spectral_data_file <- "b"
    expect_equal(.valid_mtbls_required_columns(x), character())
})

test_that(".valid_files_local works", {
    x <- MsBackendMetaboLights()
    expect_equal(.valid_files_local(x), character())
    x@spectraData <- DataFrame(a = 1:4, b = "c", dataStorage = "d")
    expect_match(.valid_files_local(x), "One or more of the data files")
})

test_that("backendMerge,MsBackendMetaboLights works", {
    ## Online mode
    be <- backendInitialize(MsBackendMetaboLights(), mtblsId = "MTBLS39",
                            filePattern = "A.cdf")
    l <- split(be, factor(be$dataOrigin, levels = unique(be$dataOrigin)))
    res <- backendMerge(l)

    expect_equal(rtime(be), rtime(res))
    expect_equal(dataOrigin(be), dataOrigin(res))
    expect_equal(mz(be), mz(res))

    ## Offline data
    a <- backendInitialize(MsBackendMetaboLights(), mtblsId = "MTBLS39",
                           filePattern = "63A.cdf", offline = TRUE)
    b <- backendInitialize(MsBackendMetaboLights(), mtblsId = "MTBLS8735",
                           filePattern = "2_E_POS.mzML")

    d <- backendMerge(a, b)
    expect_true(length(d) == (length(a) + length(b)))
    expect_equal(rtime(d), c(rtime(a), rtime(b)))
})

---
title: "Retrieve and Use Mass Spectrometry Data from MetaboLights"
output:
    BiocStyle::html_document:
        toc_float: true
vignette: >
    %\VignetteIndexEntry{Retrieve and Use Mass Spectrometry Data from MetaboLights}
    %\VignetteEngine{knitr::rmarkdown}
    %\VignetteEncoding{UTF-8}
    %\VignettePackage{MsBackendMetaboLights}
    %\VignetteDepends{Spectra,BiocStyle}
---

```{r style, echo = FALSE, results = 'asis', message=FALSE}
BiocStyle::markdown()
```

**Package**: `r Biocpkg("MsBackendMetaboLights")`<br />
**Authors**: `r packageDescription("MsBackendMetaboLights")[["Author"]] `<br />
**Last modified:** `r file.info("MsBackendMetaboLights.Rmd")$mtime`<br />
**Compiled**: `r date()`

```{r, echo = FALSE, message = FALSE}
library(Spectra)
library(BiocStyle)
```

# Introduction

The `r Biocpkg("Spectra")` package provides a central infrastructure for the
handling of Mass Spectrometry (MS) data in Bioconductor. The package supports
interchangeable use of different *backends* to import and represent MS data from
a variety of sources and data formats. The *MsBackendMetaboLights* package
allows to retrieve MS data files directly from the
[MetaboLights](https://www.ebi.ac.uk/metabolights/) repository. MetaboLights is
one of the main public repositories for deposition of metabolomics experiments
including (raw) MS and/or NMR data files and the related experimental and
analytical results. The *MsBackendMetaboLights* package downloads and locally
caches MS data files for a MetaboLights data set and enables further analyses of
this data directly in R.


# Installation

The package can be installed from within R with the commands below:

```{r, eval = FALSE}
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("RforMassSpectrometry/MsBackendMetaboLights")
```


# Importing MS Data from MetaboLights

[MetaboLights](https://www.ebi.ac.uk/metabolights/) is one of the main public
repositories for deposition of metabolomics experiments including (raw) mass
spectrometry (MS) and NMR data files and experimental/analysis results. The
experimental metadata and results are stored as plain text files in ISA-tab
format. Each MetaboLights experiment must provide a file describing the samples
analyzed and at least one *assay* file that links between the experimental
samples and the (raw and processed) data files with quantification of
metabolites/features in these samples.

In this vignette we explore and load MS data files from a small MetaboLights
experiment. MetaboLights provides information on a data set/experiment as a set
of plain text files in *ISA-tab* format. These can be accessed and read from the
data set's ftp folder. The set of files consist generally of a file with
information on the experiment/investigation (in a file with the file name
starting with *i_*) the samples of the data set (file name starting with *s_*),
the *assay* (measurements/analysis) of the experiment and a file with quantified
metabolite abundances (file name starting with *m_*). Note that a data set can
have more than one assay file.

Below we list all files from the MetaboLights data set with the ID *MTBLS39*.

```{r}
library(MsBackendMetaboLights)

#' List files of a MetaboLights data set
all_files <- mtbls_list_files("MTBLS39")
```

All these files are directly accessible in the ftp folder associated with the
MetaboLights data set. Below we use the `mtbls_ftp_path()` function to return
the ftp path for our test data set.

```{r}
mtbls_ftp_path("MTBLS39")
```

```{r, echo = FALSE}
Sys.sleep(4)
```

We could inspect the content of this folder also using a browser supporting the
ftp file transfer protocol and download individual files manually. We can
however access the files also directly from within R. Below we read the *assay*
data file directly using the base R `read.table()` function. Note that
connections to the MetaboLights FTP server are rate-limited and might thus
fail. We use below the `retry()` function from the `r Biocpkg("MsCoreUtils")`
package to retry reading from the FTP server if the connection fails or gets
closed before the data is fully read.

```{r}
#' Get the assay files of the data set
grep("^a_", all_files, value = TRUE)

#' Read the assay file
a <- MsCoreUtils::retry(
    read.table(paste0(mtbls_ftp_path("MTBLS39"),
                      grep("^a_", all_files, value = TRUE)),
               sep = "\t", header = TRUE, check.names = FALSE),
    ntimes = 5, sleep_mult = 7)
```

Each row in this assay table refers to one measurement (data file) of the data
set, with columns providing information on that measurement. The number and
content of columns can vary between data sets and depends on the information the
original researcher (manually) provided. Below we list the columns available in
the assay file of our test data set.

```{r}
colnames(a)
```

MS data files are generally provided in a column named `"Derived Spectral Data
File"` but sometimes they are also listed in a column named `"Raw Spectral Data
File"`. Note that providing MS data files is not absolutely mandatory, thus, for
some data sets no MS data files might be available. Below we list the content of
these data columns.

```{r}
a[, c("Raw Spectral Data File", "Derived Spectral Data File")]
```

For this particular data set the MS data files are provided in the `"Raw
Spectral Data File"` column. These files are in CDF format and can hence be
loaded using the `MsBackendMetaboLights` backend into R as a `Spectra` object
(`MsBackendMetaboLights` directly extends *Spectra*'s `MsBackendMzR` backend and
therefore supports import of MS data files in *mzML*, *CDF* or *mzXML*
formats). By default, all MS data files of all assays would be retrieved, but in
our example below we restrict to few data files to reduce the amount of data
that needs to be downloaded. To this end we define a pattern matching the file
name of only some data files using the `filePattern` parameter. Alternatively,
for data sets with more than one assay, it would also be possible to select MS
data files from one particular assay only using the `assayName` parameter. In
our case we load all MS data files that end with *63A.cdf*.

```{r}
library(Spectra)

#' Load MS data files of one data set
s <- Spectra("MTBLS39", filePattern = "63A.cdf",
             source = MsBackendMetaboLights())
s
```

This call now downloaded the files to the local cache and loaded these files as
a `Spectra` object. The downloading and caching of the data is handled by
Bioconductor's `r Biocpkg("BiocFileCache")`. The local cache can thus be managed
directly using functionality from that package. Any subsequent loading of the
same data files will load the locally cached versions avoiding thus repetitive
download of the same data.

Importantly, as can be seen above, the MS data files are loaded in **alphabetic order**.

The message that is shown by the call above indicates that the MS data files
were not provided in the expected column (`"Derived Spectral Data File"`) but in
the column for raw data files.

The `Spectra` object with the MS data files of the MetaboLights data set enables
now any subsequent analysis of the data in R. On top of the spectra variables
and mass peak data values that are provided by the MS data files also additional
information related to the MetaboLights data set are available as specific
*spectra variables*. We list all available spectra variables of the data set
below.

```{r}
spectraVariables(s)
```

The MetaboLights-specific variables are `"mtbls_id"`, `"mtbls_assay_name"`,
`"mtbls_assay_id"` and `"derived_spectral_data_file"` providing the
MetaboLights ID of the data set, the assay/method with which the data files
were generated and the original file path/name of the data files on the
MetaboLights ftp server.

```{r}
spectraData(s, c("mtbls_id", "mtbls_assay_name", "mtbls_assay_id",
                 "derived_spectral_data_file"))
```

These variables can be used to link the individual spectra back to the original
sample (e.g. through the *assay* and *sample* tables of the MetaboLights data
set.

The `mtbls_sync()` function can be used to *synchronize* the local content of a
`MsBackendMetaboLights`. This function checks if all data files of the backend
are available locally and eventually downloads and caches missing files.

```{r, echo = FALSE}
Sys.sleep(4)
```

```{r}
mtbls_sync(s@backend)
```

Also, it is possible to *manually* cache and download data files from
MetaboLights using the `mtbls_sync_data_files()` function. This function
evaluates if the respective data files are already cached and, if so, does not
download them again. Below we use this retrieve the local storage information on
one of the data files of the MetaboLights data set *MTBLS39*:

```{r}
res <- mtbls_sync_data_files("MTBLS39", fileName = "AM063A.cdf")
res
```

The `mtbls_cached_data_files()` function can be used to inspect and list locally
cached MetaboLights data files. This function does not require an active
internet connection since only local content is queried. With the default
settings, a `data.frame` with all available data files is returned.

```{r}
mtbls_cached_data_files()
```

Locally cached files for a MetaboLights data set can be removed using the
`mtbls_delete_cache()` function providing the ID of the MetaboLights data set
for which local data files should be removed.

# General information for a MetaboLights data set

Next to the code to cache and download MS data files, *MsBackendMetaboLights*
provides also utility functions to retrieve general data and information for a
data set from the MetaboLights repository. These functions are listed on the
`?Metabolites-utils` help page. The `mtbls_list_files()` for example allows to
list all available files in the data set's base FTP folder.

```{r}
mtbls_list_files("MTBLS39")
```

The `mtbls_assay_data()` retrieves information on the, or one of the possibly
multiple, used assay(s).

```{r}
adat <- mtbls_assay_data("MTBLS39")
head(adat)
```

The `mtbls_sample_data()` retrieves the sample information for a data set.

```{r}
sdat <- mtbls_sample_data("MTBLS39")
head(sdat)
```

And the `mtbls_metadata()` combines the sample and assay information into a
single `data.frame` from which individual sample information for the respective
MS data files could be extracted.

```{r}
mdat <- mtbls_metadata("MTBLS39")
dim(mdat)
```

# Session information

```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MsBackendMetaboLights.R
\name{MsBackendMetaboLights}
\alias{MsBackendMetaboLights}
\alias{MsBackendMetaboLights-class}
\alias{backendInitialize,MsBackendMetaboLights-method}
\alias{backendRequiredSpectraVariables,MsBackendMetaboLights-method}
\alias{mtbls_sync}
\title{MsBackend representing MS data from MetaboLights}
\usage{
MsBackendMetaboLights()

\S4method{backendInitialize}{MsBackendMetaboLights}(
  object,
  mtblsId = character(),
  assayName = character(),
  filePattern = "mzML$|CDF$|cdf$|mzXML$",
  fileName = character(),
  offline = FALSE,
  ...
)

\S4method{backendRequiredSpectraVariables}{MsBackendMetaboLights}(object, ...)

mtbls_sync(x, offline = FALSE)
}
\arguments{
\item{object}{an instance of \code{MsBackendMetaboLights}.}

\item{mtblsId}{\code{character(1)} with the ID of a single MetaboLights data
set/experiment.}

\item{assayName}{\code{character} with the file names of assay files of the data
set. If not provided (\code{assayName = character()}, the default), MS data
files of all data set's assays are loaded. Use
\verb{mtbls_list_files(<MetaboLights ID>, pattern = "^a_")} to list all
available assay files of a data set \verb{<MetaboLights ID>}.}

\item{filePattern}{\code{character} with the pattern defining the supported (or
requested) file types. Defaults to
\code{filePattern = "mzML$|CDF$|cdf$|mzXML$"} hence restricting to mzML,
CDF and mzXML files which are supported by \emph{Spectra}'s
\code{MsBackendMzR} backend.}

\item{fileName}{\code{character} with optional base file name(s) of selected
data files to load. If provided, only data files whose base file name
matches one of \code{fileName} are loaded. Defaults to
\code{fileName = character()} in which case all matching data files are
loaded. An error is thrown if none of the \code{fileName} is found in the
data set.}

\item{offline}{\code{logical(1)} whether only locally cached content should be
evaluated/loaded.}

\item{...}{additional parameters; currently ignored.}

\item{x}{an instance of \code{MsBackendMetaboLights}.}
}
\value{
\itemize{
\item For \code{MsBackendMetaboLights()}: an instance of \code{MsBackendMetaboLights}.
\item For \code{backendInitialize()}: an instance of \code{MsBackendMetaboLights} with
the MS data of the specified MetaboLights data set.
\item For \code{backendRequiredSpectraVariables()}: \code{character} with spectra
variables that are needed for the backend to provide the MS data.
\item For \code{mtbls_sync()}: the input \code{MsBackendMetaboLights} with the paths to
the locally cached data files being eventually updated.
}
}
\description{
\code{MsBackendMetaboLights} retrieves and represents mass spectrometry (MS)
data from metabolomics experiments stored in the
\href{https://www.ebi.ac.uk/metabolights/}{MetaboLights} repository. The backend
directly extends the \link[Spectra:MsBackendMzR]{Spectra::MsBackendMzR} backend from the \emph{Spectra}
package and hence supports MS data in mzML, netCDF and mzXML format. Data
in other formats can not be loaded with \code{MsBackendMetaboLights}.
Upon initialization with the \code{backendInitialize()} method, the
\code{MsBackendMetaboLights} backend downloads and caches the MS data files of
an experiment locally avoiding hence repeated download of the data. Data
files are listed and returned in \strong{alphabetic order}.
The local data cache is managed by Bioconductor's \emph{BiocFileCache} package.
See the help and vignettes from that package for details on cached data
resources. Additional utility function for management of cached files are
also provided by \emph{MsBackendMetaboLights}. See help for
\code{\link[=mtbls_cached_data_files]{mtbls_cached_data_files()}} for more information.
}
\details{
File names for data files are by default extracted from the column
\code{"Derived Spectral Data File"} of the MetaboLights data set's \emph{assay}
table. If this column does not contain any supported file names, the
assay's column \code{"Raw Spectral Data File"} is evaluated instead.

The backend uses the
\href{https://bioconductor.org/packages/BiocFileCache}{BiocFileCache} package for
caching of the data files. These are stored in the default local
\emph{BiocFileCache} cache along with additional metadata that includes the
MetaboLights ID and the assay file name with which the data file is
associated with. Note that at present only MS data files in \emph{mzML}, \emph{CDF}
and \emph{mzXML} format are supported.

The \code{MsBackendMetaboLights} backend defines and provides additional spectra
variables \code{"mtbls_id"}, \code{"mtbls_assay_name"} , \code{"mtbls_assay_id"} and
\code{"derived_spectral_data_file"} that list the MetaboLights ID, the name of
the assay file and the original data file name on the MetaboLights ftp
server for each individual spectrum. The \code{"derived_spectral_data_file"} can
be used for the mapping between the experiment's samples and the
individual data files, respective their spectra. This mapping is provided
in the MetaboLights assay file.

The \code{MsBackendMetaboLights} backend is considered \emph{read-only} and does
thus not support changing \emph{m/z} and intensity values directly.
}
\note{
To account for high server load and eventually failing or rejected
downloads from the MetaboLights ftp server, the download functions
repeatedly retry to download a file. An error is thrown if download fails
for 3 consecutive attempts. Between each attemp, the function waits
for an increasing time period (5 seconds between the first and second
and 10 seconds between the 2nd and 3rd attempt). This time period can
also be configured with the \code{"metabolights.sleep_mult"} option, which
defines the \emph{sleep time multiplicator} (defaults to 5).
}
\section{Initialization and loading of data}{


New instances of the class can be created with the \code{MsBackendMetaboLights()}
function. Data is loaded and initialized using the \code{backendInitialize()}
function which can be configured with parameters \code{mtblsId}, \code{assayName} and
\code{filePattern}. \code{mtblsId} must be the ID of a \strong{single} (existing)
MetaboLights data set. Parameter \code{assayName} allows to define specific
\emph{assays} of the MetaboLights data set from which the data files should be
loaded. If provided, it should be the file name(s) of the respective
assay(s) in MetaboLights (use e.g.
\verb{mtbls_list_files(<MetaboLights ID>, pattern = "^a_")} to list all available
assay files for a given MetaboLights ID \verb{<MetaboLights ID>}). By default,
with \code{assayName = character()} MS data files from \strong{all} assays of a data
set are loaded. Optional parameter \code{filePattern} defines the pattern that
should be used to filter the file names of the MS data files. It defaults
to data files with file endings of supported MS data files. The optional
\code{fileName} parameter allows to further restrict the loaded data to specific
data file(s), defined by their base file name(s).
\code{backendInitialize()} requires an active internet connection as the
function first compares the remote file content to the locally cached files
and eventually synchronizes changes/updates. This can be skipped with
\code{offline = TRUE} in which case only locally cached content is queried.

The \code{backendRequiredSpectraVariables()} function returns the names of the
spectra variables required for the backend to provide the MS data.

The \code{mtbls_sync()} function can be used to \emph{synchronize} the local data
cache and ensure that all data files are locally available. The function
will check the local cache and eventually download missing data files from
the MetaboLights repository.
}

\examples{

library(MsBackendMetaboLights)

## List files of a MetaboLights data set
mtbls_list_files("MTBLS39")

## Initialize a MsBackendMetaboLights representing all MS data files of
## the data set with the ID "MTBLS39". This will download and cache all
## files and subsequently load and represent them in R.

be <- backendInitialize(MsBackendMetaboLights(), "MTBLS39")
be

## The `mtbls_sync()` function can be used to ensure that all data files are
## available locally. This function will eventually download missing data
## files or update their paths.
be <- mtbls_sync(be)
}
\author{
Philippine Louail, Johannes Rainer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mtbls_utils.R
\name{MetaboLights-utils}
\alias{MetaboLights-utils}
\alias{mtbls_ftp_path}
\alias{mtbls_list_files}
\alias{mtbls_sync_data_files}
\alias{mtbls_cached_data_files}
\alias{mtbls_delete_cache}
\alias{mtbls_assay_data}
\alias{mtbls_sample_data}
\alias{mtbls_metadata}
\title{Utility functions for the MetaboLights repository}
\usage{
mtbls_ftp_path(x = character(), mustWork = TRUE)

mtbls_list_files(x = character(), pattern = NULL)

mtbls_sync_data_files(
  mtblsId = character(),
  assayName = character(),
  pattern = "mzML$|CDF$|cdf$|mzXML$",
  fileName = character()
)

mtbls_cached_data_files(
  mtblsId = character(),
  assayName = character(),
  pattern = "*",
  fileName = character()
)

mtbls_delete_cache(mtblsId = character())

mtbls_assay_data(mtblsId = character(), assayName = character())

mtbls_sample_data(mtblsId = character())

mtbls_metadata(
  mtblsId = character(),
  assayName = character(),
  keepOntology = TRUE,
  keepProtocol = TRUE,
  simplify = FALSE
)
}
\arguments{
\item{x}{\code{character(1)} with the ID of the MetaboLights data set (usually
starting with a \emph{MTBLS} followed by a number).}

\item{mustWork}{for \code{mtbls_ftp_path()}: \code{logical(1)} whether the validity of
the path should be verified or not. By default (with \code{mustWork = TRUE})
the function throws an error if either the data set does not exist or
if the folder can not be accessed (e.g. if no internet connection is
available).}

\item{pattern}{for \code{mtbls_list_files()}, \code{mtbls_sync_data_files()} and
\code{mtbls_cached_data_files()}: \code{character(1)} defining a pattern
to filter the file names, such as \code{pattern = "^a_"} to retrieve the
file names of all assay files of the data set (i.e., files with a name
starting with \code{"a_"}). This parameter is passed to the \code{\link[=grepl]{grepl()}}
function.}

\item{mtblsId}{\code{character(1)} with the ID of a \strong{single} MetaboLights data
set/experiment.}

\item{assayName}{\code{character} with the file names of assay files of the data
set. If not provided (\code{assayName = character()}, the default), MS data
files of all data set's assays are loaded. Use
\verb{mtbls_list_files(<MetaboLights ID>, pattern = "^a_")} to list all
available assay files of a data set \verb{<MetaboLights ID>}.}

\item{fileName}{for \code{mtbls_sync_data_files()} and
\code{mtbls_cached_data_files()}: optional \code{character}
defining the names of specific data files of a data set that should be
downloaded and cached.}

\item{keepOntology}{for \code{mtbls_metadata()}: \code{logical(1)} whether to keep
columns related to ontology. Default is \code{TRUE}.}

\item{keepProtocol}{for \code{mtbls_metadata()}: \code{logical(1)} whether to keep
columns with information related to protocols. Default is \code{TRUE}.}

\item{simplify}{for \code{mtbls_metadata()}: \code{logical(1)} whether to simplify
the result removing columns with only missing data or duplicated content.
Default is \code{FALSE}.}
}
\value{
\itemize{
\item For \code{mtbls_ftp_path()}: \code{character(1)} with the ftp path to the specified
data set on the MetaboLights ftp server.
\item For \code{mtbls_list_files()}: \code{character} with the names of the files in the
data set's base ftp directory.
\item For \code{mtbls_sync_data_files()} and \code{mtbls_cached_data_files()}: a
\code{data.frame} with the MetaboLights ID, the assay name(s) and remote and
local file names of the synchronized data files.
}
}
\description{
\href{https://www.ebi.ac.uk/metabolights/}{MetaboLights} is one of the main
public repositories for deposition of metabolomics experiments including
(raw) mass spectrometry (MS) and NMR data files and experimental/analysis
results. The experimental metadata and results are stored as plain text
files in ISA-tab format. Each MetaboLights experiment must provide a
file describing the samples analyzed and at least one \emph{assay} file that
links between the experimental samples and the (raw and processed) data
files with quantification of metabolites/features in these samples.

Each experiment in MetaboLights is identified with its unique identifier,
starting with \emph{MTBLS} followed by a number. The data (metadata files and
MS/NMR data files) of an experiment are available through the repository's
ftp server.

The functions listed here allow to query and retrieve information of a
data set/experiment from MetaboLights.
\itemize{
\item \code{mtbls_ftp_path()}: returns the FTP path for a provided MetaboLights ID.
With \code{mustWork = TRUE} (the default) the function throws an error if
the path is not accessible (either because the data set does not exist or
no internet connection is available). The function returns a
\code{character(1)} with the FTP path to the data set folder.
\item \code{mtbls_list_files()}: returns the available files (and directories) for the
specified MetaboLights data set (i.e., the FTP directory content of the
data set). The function returns a \code{character} vector with the relative
file names to the absolute FTP path (\code{mtbls_ftp_path()}) of the data set.
Parameter \code{pattern} allows to filter the file names and define which
file names should be returned.
\item \code{mtbls_assay_data()}: retrieves one of the \emph{assay} files for a
MetaboLights data set (parameter \code{mtblsId}) returning its content as a
\code{data.frame}. Parameter \code{assayName} allows to specify which assay file to
load (if multiple are available).
\item \code{mtbls_sample_data()}: gets the \emph{sample} file for a MetaboLights data set
(parameter \code{mtblsId}) and returns its content as a \code{data.frame}.
\item \code{mtbls_metadata()}: gets one \emph{assay} file for the specified MetaboLights
data set (parameter \code{mtblsId}) and merges it with the respective \emph{sample}
information returning the content as a \code{data.frame}. Optional parameters
\code{keepOntology}, \code{keepProtocol} and \code{simplify} allow to restrict the
returned content to fewer columns.
\item \code{mtbls_cached_data_files()}: lists locally cached data files from
MetaboLights. Since this function evaluates only local content it does not
require an internet connection. With the default parameters all available
data files are listed. The parameters can be used to restrict the lookup.
\item \code{mtbls_sync_data_files()}: synchronize data files of a specifies
MetaboLights data set eventually downloading and locally caching them.
Parameter \code{fileName} allows to specify names of selected data files to
sync.
\item \code{mtbls_delete_cache()}: removes all local content for the MetaboLights
data set with ID \code{mtblsId}. This will delete eventually present
locally cached data files for the specified data set. This does not
change any other data eventually present in the local \code{BiocFileCache}.
}
}
\details{
File names are returned in \strong{alphabetic order}.
}
\examples{

## Get the FTP path to the data set MTBLS2
mtbls_ftp_path("MTBLS2")

## Retrieve available files (and directories) for the data set MTBLS2
mtbls_list_files("MTBLS2")

## Retrieve the available assay files (file names starting with "a_").
afiles <- mtbls_list_files("MTBLS2", pattern = "^a_")
afiles

## Read the content of one file. Connections to the MetaboLights ftp server
## are limited and might fail, thus we use the `retry()` function to
## retry on failure for 5 times (waiting `i * sleep_mult` seconds in between)
a <- MsCoreUtils::retry(
    read.table(paste0(mtbls_ftp_path("MTBLS2"), afiles[1L]),
    header = TRUE, sep = "\t", check.names = FALSE),
    ntimes = 5, sleep_mult = 4)
head(a)

## Get the assay information for one MTBLS data set
mtbls_assay_data("MTBLS2")

## Get the sample information for one data set
mtbls_sample_data("MTBLS2")

## List all available files
mtbls_cached_data_files()
}
\author{
Johannes Rainer, Philippine Louail
}
